#!/usr/bin/env python
# $Id: howcome-demo.py,v 1.3 2001/09/13 15:42:03 andreas Exp $
# Find a sequence of failure-inducing program state differences,
# as described in `Isolating Cause-Effect Chains From Computer Programs'.

# Copyright
# ---------
#
# This software is copyright (C) 2000 Universitt Passau, Germany and
# (C) 2001 Universitt des Saarlandes, Germany.
#
# Permission is granted to use and modify this software FOR RESEARCH
# AND EVALUATION USAGE ONLY.  Commercial usage, usage in software
# production, and redistribution are prohibited.  All rights reserved.
#
# For technical questions, contact Andreas Zeller <zeller@acm.org>.
# Technical documentations and white papers are available from the
# Delta Debugging web site, `http://www.st.cs.uni-sb.de/dd/'.
#
# This demo version does not handle pointers or complex data
# structures.  For information on licensing HOWCOME software for
# commercial use or incorporate HOWCOME into your own environment,
# contact Andreas Zeller <zeller@acm.org>.

# How to run this program
# -----------------------
#
# To run this program, you need
# - a PC with GNU/Linux
# - Python 2.0 or later (www.python.org; Python 1.5.2 should work, too)
# - the GNU C compiler (www.gnu.org/software/gcc)
# - the GNU debugger (www.gnu.org/software/gdb)
#
# No installation is required.  To execute this program, simply run
#
#     $ python howcome-demo.py
#
# The program will
# 1. Create a file named `sample.c' in the current directory
#    (unless it already exists)
# 2. Compile `sample.c' to `sample' in the current directory
#    (unless `sample' already exists and is newer than `sample.c'
# 3. Isolate and print the cause-effect chain from `sample'.
#
# Normally, you should get lots of diagnostic output ending in:
#
#     Cause-effect chain for './sample'
#     ---------------------------------
#      
#     Arguments are 11 14 13 16 (instead of 9 8 7)
#     therefore at 'main', argc is 5 (instead of 4)
#     therefore at 'shell_sort', size is 5 (instead of 4)
#     therefore at ('sample.c', 37), a[0] is 0 (instead of 7)
#     therefore the run fails.          
#
# Feel free to adapt this program to your own test cases.  Be aware,
# though, that this demo version does not handle pointers or complex
# data structures.

# Required modules
import sys, os
import string
import re
import time
import popen2

# Measure time during existence
class TimedMessage:
    logging = 1

    def __init__(self, message):
	self.start = time.time()
	self.message = message
	self.outcome = "aborted"
	if self.logging:
	    print self.message + "..."
	    sys.stdout.flush()

    def __del__(self):
	if self.logging:
	    print self.message + "..." + self.outcome + ".  (" + \
		  '%.1f' % (time.time() - self.start) + "s)"
	    sys.stdout.flush()


# Result cache
class OutcomeCache:
    # The outcome cache is implemented as a tree.  Each node points
    # to the outcome of the remaining list.
    #
    # Example: ([1, 2, 3], PASS), ([1, 2], FAIL), ([1, 4, 5], FAIL):
    #
    #      (2, FAIL)--(3, PASS)
    #     /
    # (1, None)
    #     \
    #      (4, None)--(5, FAIL)
    
    def __init__(self):
        self.tail = {}                  # Points to outcome of tail
        self.result = None              # Result so far

    def add(self, c, result):
        """Add (C, RESULT) to the cache.  C must be a list of scalars."""
        cs = c[:]
        cs.sort()
        self._add(cs, result)

    def _add(self, c, result, start = 0):
        """Helper for ADD."""
        if start >= len(c):
            self.result = result
            return

        if not self.tail.has_key(c[start]):
            self.tail[c[start]] = OutcomeCache()
        self.tail[c[start]]._add(c, result, start + 1)

    def lookup(self, c, start = 0):
        """Return RESULT if (C, RESULT) is in the cache; None, otherwise."""
        if start >= len(c):
            return self.result
        
        if self.tail.has_key(c[start]):
            return self.tail[c[start]].lookup(c, start + 1)
        
        return None

    def lookup_superset(self, c, start = 0):
        """Return RESULT if there is some (C', RESULT) in the cache with
        C' being a superset of C or equal to C.  Otherwise, return None."""
        if start >= len(c):
            if self.result:
                return self.result
            elif self.tail != {}:
                # Select some superset
                superset = self.tail[self.tail.keys()[0]]
                return superset.lookup_superset(c, start + 1)
            else:
                return None

        if self.tail.has_key(c[start]):
            return self.tail[c[start]].lookup_superset(c, start + 1)

        # Let K0 be the largest element in TAIL such that K0 <= C[START]
        k0 = None
        for k in self.tail.keys():
            if (k0 == None or k > k0) and k <= c[start]:
                k0 = k

        if k0 != None:
            return self.tail[k0].lookup_superset(c, start)
        
        return None

    def lookup_subset(self, c, start = 0):
        """Return RESULT if there is some (C', RESULT) in the cache with
        C' being a subset of C or equal to C.  Otherwise, return None."""
        if start >= len(c):
            return self.result
        
        if self.tail.has_key(c[start]):
            return self.tail[c[start]].lookup_subset(c, start + 1)
        else:
            return self.lookup_subset(c, start + 1)



# Main Delta Debugging Algorithm
class DD:
    # Test outcomes
    PASS       = "PASS"
    FAIL       = "FAIL"
    UNRESOLVED = "UNRESOLVED"

    # Resolving directions
    ADD    = "ADD"			# Add deltas to resolve
    REMOVE = "REMOVE"			# Remove deltas to resolve

    # Debugging output (set to 1 to enable)
    debug_test      = 0
    debug_dd        = 0
    debug_split     = 0
    debug_resolve   = 0

    def __init__(self):
	self.__resolving = 0
	self.__last_reported_length = 0
        self.monotony = 0
        self.outcome_cache  = OutcomeCache()
        self.cache_outcomes = 1
        self.minimize = 1
        self.maximize = 1

    # Helpers
    def __listminus(self, c1, c2):
	"""Return a list of all elements of C1 that are not in C2."""
        s2 = {}
        for delta in c2:
            s2[delta] = 1
        
	c = []
	for delta in c1:
	    if not s2.has_key(delta):
		c.append(delta)

	return c

    def __listintersect(self, c1, c2):
	"""Return the common elements of C1 and C2."""
        s2 = {}
        for delta in c2:
            s2[delta] = 1

	c = []
	for delta in c1:
            if s2.has_key(delta):
		c.append(delta)

	return c

    def __listunion(self, c1, c2):
	"""Return the union of C1 and C2."""
        s1 = {}
        for delta in c1:
            s1[delta] = 1

	c = c1[:]
	for delta in c2:
            if not s1.has_key(delta):
		c.append(delta)

	return c

    def __listsubseteq(self, c1, c2):
        """Return 1 if C1 is a subset or equal to C2."""
        s2 = {}
        for delta in c2:
            s2[delta] = 1

        for delta in c1:
            if not s2.has_key(delta):
                return 0

        return 1

    # Output
    def coerce(self, c):
	"""Return the configuration C as a compact string"""
	# Default: use printable representation
	return `c`

    def pretty(self, c):
        """Like coerce(), but sort beforehand"""
        sorted_c = c[:]
        sorted_c.sort()
        return self.coerce(sorted_c)

    # Testing
    def test(self, c):
	"""Test the configuration C.  Return PASS, FAIL, or UNRESOLVED"""
	c.sort()

        # If we had this test before, return its result
        if self.cache_outcomes:
            cached_result = self.outcome_cache.lookup(c)
            if cached_result != None:
                return cached_result

        if self.monotony:
            # Check whether we had a passing superset of this test before
            cached_result = self.outcome_cache.lookup_superset(c)
            if cached_result == self.PASS:
                return self.PASS
            
            cached_result = self.outcome_cache.lookup_subset(c)
            if cached_result == self.FAIL:
                return self.FAIL

	if self.debug_test:
            print
	    print "test(" + self.coerce(c) + ")..."

	outcome = self._test(c)

	if self.debug_test:
	    print "test(" + self.coerce(c) + ") = " + `outcome`

        if self.cache_outcomes:
            self.outcome_cache.add(c, outcome)

	return outcome

    def _test(self, c):
	"""Stub to overload in subclasses"""
	return self.UNRESOLVED		# Placeholder


    # Splitting
    def split(self, c, n):
	"""Split C into [C_1, C_2, ..., C_n]."""
	if self.debug_split:
	    print "split(" + self.coerce(c) + ", " + `n` + ")..."

	outcome = self._split(c, n)

	if self.debug_split:
	    print "split(" + self.coerce(c) + ", " + `n` + ") = " + `outcome`

	return outcome

    def _split(self, c, n):
	"""Stub to overload in subclasses"""
	subsets = []
	start = 0
	for i in range(n):
	    subset = c[start:start + (len(c) - start) / (n - i)]
	    subsets.append(subset)
	    start = start + len(subset)
	return subsets


    # Resolving
    def resolve(self, csub, c, direction):
	"""If direction == ADD, resolve inconsistency by adding deltas
  	   to CSUB.  Otherwise, resolve by removing deltas from CSUB."""

	if self.debug_resolve:
	    print "resolve(" + `csub` + ", " + self.coerce(c) + ", " + \
		  `direction` + ")..."

	outcome = self._resolve(csub, c, direction)

	if self.debug_resolve:
	    print "resolve(" + `csub` + ", " + self.coerce(c) + ", " + \
		  `direction` + ") = " + `outcome`

	return outcome


    def _resolve(self, csub, c, direction):
	"""Stub to overload in subclasses."""
	# By default, no way to resolve
	return None


    # Test with fixes
    def test_and_resolve(self, csub, r, c, direction):
	"""Repeat testing CSUB + R while unresolved."""

	initial_csub = csub[:]

	t = self.test(csub + r)
	while t == self.UNRESOLVED:
	    self.__resolving = 1
	    csub = self.resolve(csub, c, direction)

            if csub != None and len(csub) == len(c) and direction == self.ADD:
                # Added everything: csub == c.  This has already been tested.
                csub = None

            if csub != None and len(csub) == 0 and direction == self.REMOVE:
                # Removed everything: csub == []. This has already been tested.
                csub = None

	    if csub == None:
		# Nothing left to resolve
		break

	    t = self.test(csub + r)

	self.__resolving = 0
	if csub == None:
	    return self.UNRESOLVED, initial_csub

	# assert t == self.PASS or t == self.FAIL
	return t, csub

    # Inquiries
    def resolving(self):
	"""Return 1 while resolving."""
	return self.__resolving


    # Logging
    def report_progress(self, c, title):
	if len(c) != self.__last_reported_length:
	    print
	    print title + ": " + `len(c)` + " deltas left: ", self.coerce(c)
	    self.__last_reported_length = len(c)


    # New Delta Debugging algorithm
    def dddiff(self, c):
        n = 2

	if self.debug_dd:
	    print ("dddiff(" + self.pretty(c) + ", " + `n` + ")...")

	outcome = self._dddiff([], c, n)

	if self.debug_dd:
	    print ("dddiff(" + self.pretty(c) + ", " + `n` + ") = " +
                   `outcome`)

	return outcome

    def _dddiff(self, c1, c2, n):
	run = 1
        cbar_offset = 0

	# We replace the tail recursion from the paper by a loop
	while 1:
            if self.debug_dd:
                print "dd: c1 =", self.pretty(c1)
                print "dd: c2 =", self.pretty(c2)

            assert self.test(c1) == self.PASS
            assert self.test(c2) == self.FAIL
            assert self.__listsubseteq(c1, c2)

            c = self.__listminus(c2, c1)

            if self.debug_dd:
                print "dd: c2 - c1 =", self.pretty(c)

            if n > len(c):
                # No further minimizing
                print "dd: done"
                return (c, c1, c2)

	    self.report_progress(c, "dd")

	    cs = self.split(c, n)

	    print
	    print "dd (run #" + `run` + "): trying",
	    for i in range(n):
		if i > 0:
		    print "+",
		print len(cs[i]),
	    print

            progress = 0

            next_c1 = c1[:]
            next_c2 = c2[:]
            next_n = n

	    # Check subsets
            for j in range(n):
                i = (j + cbar_offset) % n
                
                if self.debug_dd:
                    print "dd: trying", self.pretty(cs[i])

                csub = self.__listunion(c1, cs[i])
                (t, csub) = self.test_and_resolve(csub, [], c, self.REMOVE)

                if t == self.FAIL:
                    # Found
                    progress    = 1
                    next_c2     = csub
                    next_n      = 2
                    cbar_offset = 0

                    if self.debug_dd:
                        print "dd: reduce c2 to", len(next_c2), "deltas:",
                        print self.pretty(next_c2)
                    break

                if t == self.PASS:
                    # Reduce to complement
                    progress    = 1
                    next_c1     = csub
                    next_n      = max(next_n - 1, 2)
                    cbar_offset = i

                    if self.debug_dd:
                        print "dd: increase c1 to", len(next_c1), "deltas:",
                        print self.pretty(next_c1)
                    break


                csub = self.__listminus(c2, cs[i])
                (t, csub) = self.test_and_resolve(csub, [], c, self.ADD)

                if t == self.PASS:
                    # Found
                    progress    = 1
                    next_c1     = csub
                    next_n      = 2
                    cbar_offset = 0

                    if self.debug_dd:
                        print "dd: increase c1 to", len(next_c1), "deltas:",
                        print self.pretty(next_c1)
                    break

                if t == self.FAIL:
                    # Increase
                    progress    = 1
                    next_c2     = csub
                    next_n      = max(next_n - 1, 2)
                    cbar_offset = i

                    if self.debug_dd:
                        print "dd: reduce c2 to ", len(next_c2), "deltas:",
                        print self.pretty(next_c2)
                    break

            if progress:
                self.report_progress(self.__listminus(next_c2, next_c1), "dd")
            else:
                if n >= len(c):
                    # No further minimizing
                    print "dd: done"
                    return (c, c1, c2)

                next_n = min(len(c), n * 2)
                print "dd: increase granularity to", next_n
                cbar_offset = (cbar_offset * next_n) / n

            c1  = next_c1
            c2  = next_c2
            n   = next_n
	    run = run + 1

    def dd(self, c):
        return self.dddiff(c)           # Backwards compatibility


# GDB interface
class GDB:
    prompt = "(gdb)\n";
    logging = 0
    fetch_debuggee_pid = 0

    def __init__(self, debuggee):
        self.debuggee = debuggee
        self.child = popen2.Popen3("gdb " + debuggee + " 2>&1")
        self.debuggee_pid = None
        quoted_prompt = string.replace(self.prompt, "\n", "\\n")
        self.__write("set prompt " + quoted_prompt)
        initial_messages = self.__read_until_prompt()
        self.question("set editing off")
        self.question("set print demangle off")
        self.question("set confirm off")
        self.question("set verbose off")
        self.question("set language c")

        if self.fetch_debuggee_pid:
            self.question("break main")
            self.question("commands 1\ninfo program\ncont\nend")
        

    # Helpers

    # Convert a list NUMS of integers to a space-separated string
    def numlist(self, nums):
        list = ""
        for num in nums:
            if list != "":
                list = list + " "
            list = list + `num`
        return list


    # Low-level interface

    # Write CMD to GDB
    def __write(self, cmd):
        if self.logging:
            print "-> " + cmd
            sys.stdout.flush()
        self.child.tochild.write(cmd + "\n")
        self.child.tochild.flush()

    # Read reply until prompt appears
    def __read_until_prompt(self):
        reply = ""
        while reply[-len(self.prompt):] != self.prompt:
            line = self.child.fromchild.readline()
            if self.logging:
                print "<- " + line,
                sys.stdout.flush()

            if line == '':
                raise "EOF during GDB command" # FIXME: raise proper exception
                
            reply = reply + line

            if self.debuggee_pid == None and string.find(line, "Pid ") >= 0:
                pid = line[string.index(line, "Pid ") + 4:]
                pid = pid[:string.index(pid, ".")]
                self.debuggee_pid = string.atoi(pid)
            
        # Cut off prompt and final newline
        reply = reply[:-len(self.prompt)]
        return reply

    # Mid-level interface

    # Send CMD to GDB; return its reply
    def question(self, cmd):
        self.__write(cmd)
        reply = self.__read_until_prompt()
        if reply[:len(cmd)] == cmd:
            # Weed out echo
            reply = reply[len(cmd):]
            
        return reply


    # High-level interface

    # Return list of debuggee symbols
    def all_symbols(self):
        break_cmd = "break "
        symbols = string.split(self.question("complete " + break_cmd), "\n")
        for i in range(0, len(symbols)):
            symbols[i] = symbols[i][len(break_cmd):]
        return symbols

    # Return list of debuggee functions
    def all_functions(self):
        functions = []
        for s in self.all_symbols():
            # Fetch <ADDRESS> from symbol
            reply = self.question("print '" + s + "'")
            reply = reply[:string.find(reply, "\n")]
            if string.find(reply, "<") < 0:
                continue
            name = reply[string.rfind(reply, "<") + 1:]
            name = name[:string.rfind(name, ">")]
            if string.find(name, " ") >= 0 or string.find(name, ".") >= 0:
                continue
            functions.append(name)

        return functions

    # Break at LOCATION (a function, a line number, or a (FILE, LINE) pair);
    # return breakpoint number
    def break_at(self, location):
        if isinstance(location, type("")):
            location = "'" + location + "'"
        elif isinstance(location, type(0)):
            location = `location`
        elif isinstance(location, type(("", 0))):
            (file, line) = location
            location = file + ":" + `line`
        
        reply = self.question("break " + location)

        # REPLY is something like `Breakpoint NR at ADDRESS'
        info = reply[string.index(reply, "Breakpoint "):]
        return string.atoi(string.split(info)[1])

    # Define sequence of commands for given breakpoint
    def commands(self, breakpoint, cmds):
        cmd = "commands " + `breakpoint` + "\n"
        for c in cmds:
            cmd = cmd + c + "\n"
        cmd = cmd + "end"
        self.question(cmd)

    # Enable the given breakpoints
    def enable(self, breakpoints):
        if len(breakpoints) > 0:
            self.question("enable " + self.numlist(breakpoints))

    # Disable the given breakpoints
    def disable(self, breakpoints):
        if len(breakpoints) > 0:
            self.question("disable " + self.numlist(breakpoints))

    # Delete the given breakpoints
    def delete(self, breakpoints):
        if len(breakpoints) > 0:
            self.question("delete " + self.numlist(breakpoints))

    # Interactive mode. Useful for debugging.
    def interactive(self):
        saved_logging = self.logging
        self.logging = 0
        resume = "#"
        print "Entering interactive mode.  Enter " + `resume` + " to resume."
        while 1:
            sys.stdout.write("(gdb) ")
            sys.stdout.flush()

            cmd = sys.stdin.readline()
            cmd = cmd[:len(cmd) - 1]
            if cmd == resume:
                break
            sys.stdout.write(self.question(cmd))
        
        self.logging = saved_logging



# GDB interface with special state fetching capability
class StateGDB(GDB):
    CHAR_PATTERN    = re.compile("(?P<value>\d+) +(?P<character>'.*')")
    STRING_PATTERN  = re.compile('(?P<pointer>0x[^ ]+) +(?P<string>".*")')
    POINTER_PATTERN = re.compile(".*(?P<value>0x[0-9a-f]+)")
    FUNCTION_POINTER_PATTERN = re.compile(".*(?P<value>0x[0-9a-f]+) <(?P<identifier>[^>]*)>")

    def _unfold_pointer(self, name, frame, value, vars):
        if string.find(value, "(void *)") >= 0:
            return                  # Generic pointer - ignore

        if value == "0x0":
            vars[(name, frame)] = "0x0" # NULL pointer
            return

        deref_value = self.question("output *(" + name + ")")

        if deref_value[:2] == '{\n':
            # We have a pointer structure.  Dereference elements.
            SEP = " = "
            for line in string.split(deref_value, "\n"):
                separator = string.find(line, SEP)
                if separator < 0:
                    continue

                member_name = name + "->" + line[0:separator]
                member_value = line[separator + len(SEP):]

                self._fetch_values(member_name, frame, member_value, vars)

            return

        if deref_value[0] == '{':
            # Some function pointer.  Leave it unchanged.
            return
            
        # Otherwise, assume it's an array or object on the heap
        # (hopefully)

        if name == "argv":
            # Special handling for program arguments.
            elems = 4096
        else:
            # This trick will get the number of elements (if we use
            # GNU malloc on Linux)
            elems = self.question("output (((int *)(" + name +
                                  "))[-1] & ~0x1) /" +
                                  "sizeof(*(" + name + "))")

            elems = string.atoi(elems)

        if elems > 10000:
            return                      # Cannot handle this

        beyond = 1                      # Look 1 elements beyond bounds

        for i in range(0, elems + beyond):
            elem_name = name + "[" + `i` + "]"
            elem_value = self.question("output " + elem_name)

            self._fetch_values(elem_name, frame, elem_value, vars)
            if name == "argv" and elem_value == "0x0":
                break               # Stop it


    # Fetch a value NAME = VALUE from GDB into VARS, with special
    # handling of pointer structures.
    def _fetch_values(self, name, frame, value, vars):
        value = string.strip(value)
        value = string.replace(value, "\n", "")
        
        # print "Handling " + name + " = " + value

        # GDB reports characters as VALUE 'CHARACTER'.  Prefer CHARACTER.
        m = self.CHAR_PATTERN.match(value)
        if m != None:
            vars[(name, frame)] = m.group('character')
            return
        
        # GDB reports strings as POINTER "STRING".  Prefer STRING.
        m = self.STRING_PATTERN.match(value)
        if m != None:
            vars[(name, frame)] = m.group('string')
            return

        # GDB reports function pointers as POINTER <IDENTIFIER>.
        # Prefer IDENTIFIER.
        m = self.FUNCTION_POINTER_PATTERN.match(value)
        if m != None:
            vars[(name, frame)] = m.group('identifier')
            return

        # In case of pointers, unfold the given data structure
        if self.POINTER_PATTERN.match(value):
            self._unfold_pointer(name, frame, value, vars)
            return

        # Anything else: Just store the value
        vars[(name, frame)] = value

    
    # Store mapping (variable, frame) => values in VARS
    def _fetch_variables(self, frame, vars):
        SEP = " = "
        IDENTIFIER = re.compile("[a-zA-Z_]")
        for query in ["info locals", "info args"]:
            list = self.question(query)
            lines = string.split(list, "\n")

            # Some values as reported by GDB are split across several lines
            for i in range(1, len(lines)):
                if lines[i] != "" and not IDENTIFIER.match(lines[i][0]):
                    lines[i - 1] = lines[i - 1] + string.strip(lines[i])
                    
            for line in lines:
                separator = string.find(line, SEP)
                if separator > 0:
                    name  = line[0:separator]
                    value = line[separator + len(SEP):]
                    self._fetch_values(name, frame, value, vars)

        return vars

        
    # Return mapping (variable, frame) => values
    def state(self):
        t = TimedMessage("Capturing state")
        vars = {}
        frame = 0
        reply = "#0"
        while string.find(reply, "#") != -1:
            self._fetch_variables(frame, vars)
            reply = self.question("up")
            frame = frame + 1

        t.outcome = `len(vars.keys())` + " variables"
        return vars

    # Return (opaque) list of deltas
    def deltas(self, state_1, state_2):
        deltas = []
        for var in state_1.keys():
            if not state_2.has_key(var):
                continue                # Uncomparable

            value_1 = state_1[var]
            value_2 = state_2[var]
            if value_1 != value_2:
                deltas.append((var, value_1, value_2))

        deltas.sort()
        return deltas


    # Apply (opaque) list of deltas to STATE_1
    def apply_deltas(self, deltas):
        cmds = self.delta_cmds(deltas)
        for cmd in cmds:
            self.question(cmd)

    def delta_cmds(self, deltas):
        cmds = []
        current_frame = None
        for delta in deltas:
            (var, value_1, value_2) = delta
            (name, frame) = var
            if frame != current_frame:
                cmds.append("frame " + `frame`)
                current_frame = frame
                
            cmds.append("set variable " + name + " = " + value_2)

        return cmds


# Variant of DD applied to program states
class StateDD(DD):
    debug_state = 0

    def __init__(self, debuggee, args_1, args_2, tester):
        DD.__init__(self)
        self.gdb = StateGDB(debuggee)

        self.args_1 = args_1
        self.args_2 = args_2
        self.tester = tester

    def set_syncpoint(self, location):
        t = TimedMessage("Comparing states at " + `location`)
        self.syncpoint = self.gdb.break_at(location)

        self.gdb.question("set args " + self.args_1)
        self.gdb.question("run")
        self.state_1 = self.gdb.state()
        if self.debug_state:
            print "state_1 =", self.state_1

        self.gdb.question("set args " + self.args_2)
        self.gdb.question("run")
        self.state_2 = self.gdb.state()
        if self.debug_state:
            print "state_2 =", self.state_2

        self.deltas = self.gdb.deltas(self.state_1, self.state_2)

        if self.debug_state:
            serial = 1
            for delta in self.deltas:
                (var, value_1, value_2) = delta
                (name, frame) = var
                print "Delta #" + `serial` + " at frame " + `frame` + ":"
                print `name` + " = " + value_1 + " / " + value_2
                serial = serial + 1

        self.gdb.disable([self.syncpoint])
        t.outcome = `len(self.deltas)` + " deltas"

    def all_deltas(self):
        return self.deltas

    def setup(self):
        self.gdb.enable([self.syncpoint])
        self.gdb.question("set args " + self.args_1)
        return self.gdb.question("run")

    def cont(self):
        self.gdb.disable([self.syncpoint])
        return self.gdb.question("cont")

    def _test(self, c):
        self.setup()
        self.gdb.apply_deltas(c)
        final_output = self.cont()

        return self.tester(final_output)

    def cause_effect_chain(self, syncpoints):
        t1 = TimedMessage("Computing cause-effect chain")
        causes = {}
        for syncpoint in syncpoints:
            print
            print "At", `syncpoint`
            print "---" + "-" * len(`syncpoint`)
            print
            self.set_syncpoint(syncpoint)

            t2 = TimedMessage("Isolating failure-inducing deltas for " +
                             `syncpoint`)
            (c, c1, c2) = self.dd(self.all_deltas())
            causes[syncpoint] = c
            t2.outcome = `len(c)` + " deltas"
            
            del t2
            print

        t1.outcome = "done"
        return causes

    def print_cause_effect_chain(self, syncpoints):
        causes = self.cause_effect_chain(syncpoints)

        print
        print
        print "Cause-effect chain for " + `self.gdb.debuggee`
        print "=======================" + "=" * len(`self.gdb.debuggee`)
        print

        print "Arguments are " + self.args_2,
        print "(instead of " + self.args_1 + ")"

        for syncpoint in syncpoints:
            print "therefore at " + `syncpoint` + ",",

            for delta in causes[syncpoint]:
                (var, value_1, value_2) = delta
                (name, frame) = var
                print name, "=", value_2, "(instead of", value_1 + ")"
                
        print "therefore the run fails."


def sample_test():
    def create_sample():
        try:
            open("sample.c")
            return
        except:
            pass
        
        t = TimedMessage("Creating sample.c")
        open("sample.c", "w").write(
'''/* sample.c -- Sample C program to be debugged with DDD */

#include <stdio.h>
#include <stdlib.h>

static void shell_sort(int a[], int size)
{
    int i, j;
    int h = 1;
    do {
        h = h * 3 + 1;
    } while (h <= size);
    do {
        h /= 3;
        for (i = h; i < size; i++)
        {
            int v = a[i];
            for (j = i; j >= h && a[j - h] > v; j -= h)
                a[j] = a[j - h];
            if (i != j)
                a[j] = v;
        }
    } while (h != 1);
}

int main(int argc, char *argv[])
{
    int *a;
    int i;

    a = (int *)malloc((argc - 1) * sizeof(int));
    for (i = 0; i < argc - 1; i++)
        a[i] = atoi(argv[i + 1]);

    shell_sort(a, argc);

    printf("Output: ");
    for (i = 0; i < argc - 1; i++)
        printf("%d ", a[i]);
    printf("\\n");

    free(a);

    return 0;
}
''')
        t.outcome = "done"

    def compile_sample():
        try:
            # Recompile if needed
            sample_c_mtime = os.path.getmtime("sample.c")
            sample_mtime = os.path.getmtime("sample")
            compile = (sample_c_mtime > sample_mtime)
        except:
            compile = 1

        if not compile:
            return

        t = TimedMessage("Compiling sample.c")
        try:
            os.unlink("sample")
        except:
            pass
        
        for compiler in ["gcc", "cc"]:
            os.system(compiler + " -g -o sample sample.c")
            if open("sample"):
                break
        t.outcome = "done"

    def sample_tester(final_output):
        if string.find(final_output, " 0") >= 0:
            return DD.FAIL
        if string.find(final_output, "7 8 9") >= 0:
            return DD.PASS
        return DD.UNRESOLVED


    # Main function begins here
    assert os.uname()[0] == "Linux"
    
    create_sample()
    compile_sample()

    StateGDB.logging = 1
    StateDD.debug_state = 1
    StateDD.debug_test = 1

    t = TimedMessage("Starting GDB")
    dd = StateDD("./sample", "9 8 7", "11 14", sample_tester)
    t.outcome = "done"
    del t
    
    dd.print_cause_effect_chain(["main", "shell_sort", ("sample.c", 37)])


if __name__ == '__main__':
    sample_test()
